@@ -24,6 +24,10 @@ use lazy_regex::{regex_captures, regex_replace};
 use std::cmp::Ordering;
 use std::str::FromStr;
 
+pub mod upstream;
+pub mod vcs;
+pub mod vendor;
+
 /// A Debian version string
 ///
 ///
@@ -83,6 +87,14 @@ fn test_non_digit_cmp() {
     assert_eq!(non_digit_cmp("~", "a"), Ordering::Less);
 }
 
+fn drop_leading_zeroes(mut s: &str) -> &str {
+    // Drop leading zeroes while the next character is a digit
+    while s.starts_with('0') && s.chars().nth(1).map_or(false, |c| c.is_ascii_digit()) {
+        s = &s[1..];
+    }
+    s
+}
+
 fn version_cmp_part(mut a: &str, mut b: &str) -> Ordering {
     while !a.is_empty() || !b.is_empty() {
         // First, create a for the non-digit leading part of each string
@@ -148,12 +160,12 @@ impl Ord for Version {
             return std::cmp::Ord::cmp(&self_norm.0, &other_norm.0);
         }
 
-        match version_cmp_part(self_norm.1.as_str(), other_norm.1.as_str()) {
+        match version_cmp_part(self_norm.1, other_norm.1) {
             Ordering::Equal => (),
             ordering => return ordering,
         }
 
-        version_cmp_part(self_norm.2.as_str(), other_norm.2.as_str())
+        version_cmp_part(self_norm.2, other_norm.2)
     }
 }
 
@@ -212,16 +224,16 @@ impl FromStr for Version {
     }
 }
 
-impl ToString for Version {
-    fn to_string(&self) -> String {
-        let mut ret = vec![self.upstream_version.clone()];
+impl std::fmt::Display for Version {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
         if let Some(epoch) = self.epoch.as_ref() {
-            ret.insert(0, format!("{}:", epoch));
+            write!(f, "{}:", epoch)?;
         }
+        f.write_str(&self.upstream_version)?;
         if let Some(debian_revision) = self.debian_revision.as_ref() {
-            ret.push(format!("-{}", debian_revision));
+            write!(f, "-{}", debian_revision)?;
         }
-        ret.concat()
+        Ok(())
     }
 }
 
@@ -241,11 +253,11 @@ impl Version {
     ///
     /// This will return an explicit 0 for epochs and debian revisions
     /// that are not set.
-    fn explicit(&self) -> (u32, String, String) {
+    fn explicit(&self) -> (u32, &str, &str) {
         (
             self.epoch.unwrap_or(0),
-            self.upstream_version.trim_start_matches('0').to_string(),
-            self.debian_revision.as_deref().unwrap_or("0").to_string(),
+            self.upstream_version.as_str(),
+            self.debian_revision.as_deref().unwrap_or("0")
         )
     }
 
@@ -264,18 +276,20 @@ impl Version {
             Some(epoch) => Some(epoch),
         };
 
-        let upstream_version = self.upstream_version.trim_start_matches('0');
+        let mut upstream_version = self.upstream_version.as_str();
+
+        upstream_version = drop_leading_zeroes(upstream_version);
 
         let debian_revision = match self.debian_revision.as_ref() {
             Some(r) if r.chars().all(|c| c == '0') => None,
             None => None,
-            Some(revision) => Some(revision.clone()),
+            Some(revision) => Some(drop_leading_zeroes(revision)),
         };
 
         Version {
             epoch,
             upstream_version: upstream_version.to_string(),
-            debian_revision,
+            debian_revision: debian_revision.map(|r| r.to_string()),
         }
     }
 
@@ -328,30 +342,42 @@ mod tests {
             "0:1.0-2".parse::<Version>().unwrap().canonicalize(),
             "1.0-2".parse::<Version>().unwrap()
         );
+        assert_eq!(
+            "0001.0-0".parse::<Version>().unwrap().canonicalize(),
+            "1.0".parse::<Version>().unwrap()
+        );
+        assert_eq!(
+            "000.1".parse::<Version>().unwrap().canonicalize(),
+            "0.1".parse::<Version>().unwrap()
+        );
     }
 
     #[test]
     fn test_explicit() {
         assert_eq!(
-            (0, "1.0".to_string(), "1".to_string()),
+            (0, "1.0", "1"),
             "1.0-1".parse::<Version>().unwrap().explicit()
         );
         assert_eq!(
-            (1, "1.0".to_string(), "1".to_string()),
+            (1, "1.0", "1"),
             "1:1.0-1".parse::<Version>().unwrap().explicit()
         );
         assert_eq!(
-            (0, "1.0".to_string(), "0".to_string()),
+            (0, "1.0", "0"),
             "1.0".parse::<Version>().unwrap().explicit()
         );
         assert_eq!(
-            (0, "1.0".to_string(), "0".to_string()),
+            (0, "1.0", "0"),
             "1.0-0".parse::<Version>().unwrap().explicit()
         );
         assert_eq!(
-            (1, "1.0".to_string(), "0".to_string()),
+            (1, "1.0", "0"),
             "1:1.0-0".parse::<Version>().unwrap().explicit()
         );
+        assert_eq!(
+            (0, "000.1", "0"),
+            "000.1".parse::<Version>().unwrap().explicit()
+        );
     }
 
     macro_rules! assert_cmp(
@@ -363,6 +389,8 @@ mod tests {
     #[test]
     fn test_version_cmp_part() {
         assert_eq!(version_cmp_part("1.0", "1.0"), Ordering::Equal);
+        assert_eq!(version_cmp_part("0.1", "0.1"), Ordering::Equal);
+        assert_eq!(version_cmp_part("000.1", "0.1"), Ordering::Equal);
         assert_eq!(version_cmp_part("1.0", "2.0"), Ordering::Less);
         assert_eq!(version_cmp_part("1.0", "0.0"), Ordering::Greater);
         assert_eq!(version_cmp_part("10.0", "2.0"), Ordering::Greater);
@@ -396,6 +424,9 @@ mod tests {
         assert_cmp!("1.0a-1", "1.0-1", Greater);
         assert_cmp!("1.0-1", "1.0a-1", Less);
         assert_cmp!("1.0a-1", "1.0a-1", Equal);
+
+        // Bug 27
+        assert_cmp!("23.13.9-7", "0.6.45-2", Greater);
     }
 
     #[test]
@@ -503,6 +534,15 @@ mod tests {
             }
             .to_string()
         );
+        assert_eq!(
+            "1:1.0",
+            Version {
+                epoch: Some(1),
+                upstream_version: "1.0".to_string(),
+                debian_revision: None,
+            }
+            .to_string()
+        );
     }
 
     #[test]
@@ -553,7 +593,10 @@ impl sqlx::Type<Postgres> for Version {
 
 #[cfg(feature = "sqlx")]
 impl sqlx::Encode<'_, Postgres> for Version {
-    fn encode_by_ref(&self, buf: &mut sqlx::postgres::PgArgumentBuffer) -> sqlx::encode::IsNull {
+    fn encode_by_ref(
+        &self,
+        buf: &mut sqlx::postgres::PgArgumentBuffer,
+    ) -> Result<sqlx::encode::IsNull, Box<dyn std::error::Error + Send + Sync>> {
         sqlx::Encode::<Postgres>::encode_by_ref(&self.to_string().as_str(), buf)
     }
 }
@@ -585,10 +628,10 @@ use pyo3::prelude::*;
 
 #[cfg(feature = "python-debian")]
 impl FromPyObject<'_> for Version {
-    fn extract(ob: &PyAny) -> PyResult<Self> {
-        let debian_support = Python::import(ob.py(), "debian.debian_support")?;
+    fn extract_bound(ob: &Bound<PyAny>) -> PyResult<Self> {
+        let debian_support = Python::import_bound(ob.py(), "debian.debian_support")?;
         let version_cls = debian_support.getattr("Version")?;
-        if !ob.is_instance(version_cls)? {
+        if !ob.is_instance(&version_cls)? {
             return Err(pyo3::exceptions::PyTypeError::new_err("Expected a Version"));
         }
         Ok(Version {
@@ -605,7 +648,7 @@ impl FromPyObject<'_> for Version {
 #[cfg(feature = "python-debian")]
 impl ToPyObject for Version {
     fn to_object(&self, py: Python) -> PyObject {
-        let debian_support = py.import("debian.debian_support").unwrap();
+        let debian_support = py.import_bound("debian.debian_support").unwrap();
         let version_cls = debian_support.getattr("Version").unwrap();
         version_cls
             .call1((self.to_string(),))
@@ -629,15 +672,15 @@ mod python_tests {
         use pyo3::prelude::*;
 
         Python::with_gil(|py| {
-            let globals = pyo3::types::PyDict::new(py);
+            let globals = pyo3::types::PyDict::new_bound(py);
             globals
                 .set_item(
                     "debian_support",
-                    py.import("debian.debian_support").unwrap(),
+                    py.import_bound("debian.debian_support").unwrap(),
                 )
                 .unwrap();
             let v = py
-                .eval("debian_support.Version('1.0-1')", Some(globals), None)
+                .eval_bound("debian_support.Version('1.0-1')", Some(&globals), None)
                 .unwrap()
                 .extract::<Version>()
                 .unwrap();
@@ -666,7 +709,7 @@ mod python_tests {
             let v = v.to_object(py);
             let expected: Version = "1:1.0-1".parse().unwrap();
             assert_eq!(v.extract::<Version>(py).unwrap(), expected);
-            assert_eq!(v.as_ref(py).get_type().name().unwrap(), "Version");
+            assert_eq!(v.bind(py).get_type().name().unwrap(), "Version");
         });
     }
 }
@@ -692,6 +735,7 @@ impl<'de> serde::Deserialize<'de> for Version {
     }
 }
 
+/// Trait for converting an argument into a Version
 pub trait AsVersion {
     fn as_version(self) -> Result<Version, ParseError>;
 }
